package be.cylab.java.wowa.training;


import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.logging.Level;

/**
 * Class Trainer. Learn wowa weights based on training dataset
 */
public class Trainer {

    //Parameters for the trainer
    private final TrainerParameters parameters;

    //Factory to create to correct type of Solution during execution
    private final Factory factory;

    /**
     * Trainer constructor.
     *
     * @param parameters TrainerParameters Parameters for trainer
     */
    public Trainer(final TrainerParameters parameters,
                   final AbstractSolution solution_type) {
        this.parameters = parameters;
        this.factory = new Factory(solution_type);
    }

    /**
     * Run method with file as arguments.
     *
     * @param data_file_name
     * @param expected_file_name
     * @return
     */
    public final AbstractSolution run(
            final String data_file_name,
            final String expected_file_name) {
        List<List<Double>> data
                = Utils.convertJsonToDataForTrainer(data_file_name);
        List<Double> expected
                = Utils.convertJsonToExpectedForTrainer(expected_file_name);
        return this.run(data, expected);
    }

    /**
     * @param data
     * @param expected
     * @return
     */
    public final AbstractSolution run(
            final List<List<Double>> data,
            final List<Double> expected) {

        List<AbstractSolution> current_population
                = this.generateInitialPopulationAndComputeDistances(
                this.getParameters().getPopulationSize(),
                data,
                expected);

        AbstractSolution best_solution
                = this.findBestSolution(current_population);
        for (int generation = 0;
             generation < this.parameters.getMaxGenerationNumber();
             generation++) {
            current_population = this.performReproduction(
                    current_population, data, expected);
            AbstractSolution best_solution_of_current_population
                    = this.findBestSolution(current_population);

            //Log best element of the generation
            if (this.getParameters().getLogger() != null) {
                this.getParameters().getLogger().log(
                        Level.INFO,
                        "Generation " + (generation + 1) + " : "
                                + best_solution.toString());
            }

            //We found a new best solution
            if (best_solution_of_current_population.getFitnessScore()
                    < best_solution.getFitnessScore()) {
                best_solution = best_solution_of_current_population;
            }

        }
        return best_solution;
    }

    /**
     * Method to perform a k-fold cross validation.
     *
     * @param data
     * @param expected
     * @param fold_number
     * @return
     */
    public final HashMap<AbstractSolution, Double> runKFold(
            final List<List<Double>> data,
            final List<Double> expected,
            final int fold_number,
            final int increase_ration_alert) {
        TrainingDataset dataset = new TrainingDataset(data, expected);
        List<TrainingDataset> folds = prepareFolds(dataset, fold_number);
        HashMap<AbstractSolution, Double> map = new HashMap<>();
        for (int i = 0; i < fold_number; i++) {
            TrainingDataset testing = folds.get(i);
            TrainingDataset learning = new TrainingDataset();
            for (int j = 0; j < fold_number; j++) {
                if (j != i) {
                    learning.addFoldInDataset(folds, j);
                }
            }
            TrainingDataset dataset_increased = increaseTrueAlert(
                    learning.getData(),
                    learning.getExpected(),
                    increase_ration_alert);
            AbstractSolution sol = run(
                    dataset_increased.getData(),
                    dataset_increased.getExpected());
            Double score = sol.computeAUC(
                    testing.getData(),
                    testing.getExpected());

            map.put(sol, score);
        }
        return map;

    }

    /**
     * Method to perform a cross validation with filename as argument.
     *
     * @param data_file_name
     * @param expected_file_name
     * @param fold_number
     * @return
     */
    public final HashMap<AbstractSolution, Double> runKFold(
            final String data_file_name,
            final String expected_file_name,
            final int fold_number,
            final int increase_ratio_alert) {
        List<List<Double>> data
                = Utils.convertJsonToDataForTrainer(data_file_name);
        List<Double> expected
                = Utils.convertJsonToExpectedForTrainer(expected_file_name);
        return runKFold(data, expected, fold_number, increase_ratio_alert);
    }

    /**
     * Method to increase the number of true alert in data.
     * to increase penalty to do not detect a true alert
     *
     * @param data
     * @param expected
     * @param increase_ratio
     */
    final TrainingDataset increaseTrueAlert(
            final List<List<Double>> data,
            final List<Double> expected,
            final int increase_ratio) {
        int data_size = expected.size();
        for (int i = 0; i < data_size; i++) {
            if (expected.get(i) == 1) {
                for (int j = 0; j < increase_ratio; j++) {
                    expected.add(expected.get(i));
                    data.add(data.get(i));
                }
            }
        }
        return new TrainingDataset(data, expected);
    }

    /**
     * Method to separate randomly the base dataset in X folds dataset.
     *
     * @param dataset
     * @param fold_number
     * @return
     */
     final List<TrainingDataset> prepareFolds(
            final TrainingDataset dataset,
            final int fold_number) {
        //List<List<Double>> data = dataset.getData();
        List<Double> expected = dataset.getExpected();
        List<TrainingDataset> fold_dataset = new ArrayList<>();
        //Check if it is rounded !!!!
        int alert_number
                = (int) Math.floor(Utils.sumListElements(expected)
                / fold_number);
        int no_alert_number = (int) (expected.size()
                - Utils.sumListElements(expected)) / fold_number;

        for (int i = 0; i < fold_number; i++) {
            TrainingDataset tmp = new TrainingDataset();
            int alert_counter = 0;
            int no_alert_counter = 0;
            while (tmp.getLength() < (alert_number + no_alert_number)) {
                int index = Utils.randomInteger(0, dataset.getLength() - 1);
                if (dataset.getExpected().get(index) == 1
                        && alert_counter < alert_number) {
                    tmp.addElementInDataset(dataset, index);
                    dataset.removeElementInDataset(index);
                    alert_counter++;
                } else if (dataset.getExpected().get(index) == 0
                        && no_alert_counter < no_alert_number) {
                    tmp.addElementInDataset(dataset, index);
                    dataset.removeElementInDataset(index);
                    no_alert_counter++;
                }
            }
            fold_dataset.add(tmp);
        }
        return fold_dataset;
    }

    /**
     * Find the best element in the population based on its fitness score.
     *
     * @param solutions
     * @return
     */
    final AbstractSolution findBestSolution(
            final List<AbstractSolution> solutions) {

        AbstractSolution best_solution = solutions.get(0);

        for (AbstractSolution solution : solutions) {
            if (solution.getFitnessScore() < best_solution.getFitnessScore()) {
                best_solution = solution;
            }
        }
        return best_solution;
    }

    /**
     * Generate the initial population.
     *
     * @param number_of_weights
     * @param population_size
     * @return
     */
    final List<AbstractSolution> generateInitialPopulation(
            final int number_of_weights,
            final int population_size) {

        List<AbstractSolution> population = new ArrayList<>();
        for (int i = 0; i < population_size; i++) {
            AbstractSolution solution
                    = this.factory.createSolutionObject(number_of_weights);
            population.add(solution);
        }
        return population;
    }

    /**
     * This method generates an initial population with specific element.
     * A part of the population is random (as generateInitialPopulation)
     * The last elements are specific elements like [0, 0, 1, 0], [1, 0, 0, 0].
     * All weights in a vector are null except one that is equal to 1.
     * There are number_of_weight^2 specific elements or half of the population
     * Depends on the population size and the number of weight
     *
     * @param number_of_weight
     * @param population_size
     * @return
     */
    final List<AbstractSolution> generateQuasiRandomInitialPopulation(
            final int number_of_weight,
            final int population_size) {
        //Compute the number of random element in the population
        int number_of_random_element = population_size / 2;
        if (Math.pow(number_of_weight, 2) < population_size / (double) 2) {
            number_of_random_element =
                    (int) (population_size - Math.pow(number_of_weight, 2));
        }
        //Random part of the population
        List<AbstractSolution> population = this.generateInitialPopulation(
                number_of_weight,
                number_of_random_element);

        //Set specific elements in the population.
        //random_memory is used to save which specific element is already set
        int[][] random_memory = new int[number_of_weight][number_of_weight];
        while (population.size() < population_size) {
            int position_w = Utils.randomInteger(0, number_of_weight - 1);
            int position_p = Utils.randomInteger(0, number_of_weight - 1);
            if (random_memory[position_w][position_p] == 0) {
                AbstractSolution solution
                        = this.factory.createSolutionObject(number_of_weight);
                List<Double> w
                        = Utils.initializeListWithZeroValues(number_of_weight);
                List<Double> p
                        = Utils.initializeListWithZeroValues(number_of_weight);
                w.set(position_w, 1.0);
                p.set(position_p, 1.0);
                solution.setWeightsW(w);
                solution.setWeightsP(p);
                population.add(solution);
                random_memory[position_w][position_p] = 1;
            }
        }
        return population;

    }

    /**
     * @param solutions
     * @param data
     * @param expected
     * @return
     */
    final List<AbstractSolution> computeDistances(
            final List<AbstractSolution> solutions,
            final List<List<Double>> data,
            final List<Double> expected) {

        for (AbstractSolution solution : solutions) {
            solution.computeScoreTo(data, expected);
        }
        return solutions;

    }

    /**
     * Select elements for the next generation.
     * Each element has a probability to be chosen proportional to its
     * fitness score (distance for example)
     *
     * @param solutions
     * @param selected_elements
     * @param count
     * @return
     */
    final List<AbstractSolution> rouletteWheelSelection(
            final List<AbstractSolution> solutions,
            final List<AbstractSolution> selected_elements,
            final int count) {

        double max = Utils.findMaxDistance(solutions);
        double min = Utils.findMinDistance(solutions);
        double sum = Utils.sumTotalDistance(solutions);

        if (solutions.size() < count) {
            throw new IllegalArgumentException(
                    "Not enough elements in population to select "
                            + count + "parents"
            );
        }

        while (selected_elements.size() < count) {
            double tos = Math.random();
            double normalized_distance = 0;
            for (AbstractSolution solution : solutions) {
                normalized_distance = normalized_distance
                        + (max + min - solution.getFitnessScore()) / sum;
                if (normalized_distance > tos) {
                    selected_elements.add(solution);
                    solutions.remove(solution);
                    break;
                }
            }
        }
        return selected_elements;
    }

    /**
     * Select elements for the next generation.
     * Two elements are randomly selected. The element with the best
     * fitness score is selected. The other is put again in the population
     *
     * @param solutions
     * @param selected_elements
     * @param count
     * @return
     */
    final List<AbstractSolution> tournamentSelection(
            final List<AbstractSolution> solutions,
            final List<AbstractSolution> selected_elements,
            final int count) {

        while (selected_elements.size() < count) {

            //Select random element in the population
            int solution1_index = Utils.randomInteger(0, solutions.size() - 1);
            AbstractSolution solution1 = solutions.get(solution1_index);

            //Select random element in the population
            int solution2_index = Utils.randomInteger(0, solutions.size() - 1);
            AbstractSolution solution2 = solutions.get(solution2_index);

            //Compare the two selected element and put the solution with
            // the smallest distance in selected list
            if (solution1.getFitnessScore() < solution2.getFitnessScore()) {
                selected_elements.add(solution1);
                solutions.remove(solution1);
            } else {
                selected_elements.add(solution2);
                solutions.remove(solution2);

            }
        }
        return selected_elements;
    }

    /**
     * Method to select parents for the mating.
     * Use an elitism mechanism. The two best elements of the current population
     * are systematically selected for the next one.
     *
     * @param solutions
     * @param count
     * @param selection_method
     * @return
     */
    final List<AbstractSolution> selectParents(
            final List<AbstractSolution> solutions,
            final int count,
            final int selection_method) {

        List<AbstractSolution> selected_parents = new ArrayList<>();
        //Select the two best current solutions

        //Sort population
        Collections.sort(solutions);
        //Put the two best elements in selected parents list (elitism)
        selected_parents.add(solutions.get(0));
        selected_parents.add(solutions.get(1));
        //Remove the two best elements from the list solutions
        // to avoid select same elements several times
        solutions.remove(0);
        solutions.remove(1);

        if (selection_method == TrainerParameters.SELECTION_METHOD_RWS) {

            return this.rouletteWheelSelection(solutions,
                    selected_parents,
                    count
            );
        } else if (selection_method == TrainerParameters.SELECTION_METHOD_TOS) {
            return this.tournamentSelection(solutions,
                    selected_parents,
                    count
            );
        }
        throw new IllegalArgumentException("Invalid selection method");

    }

    /**
     * Select two random elements in the selected parents population.
     * Reproduce these two parents to generate 2 news individuals
     *
     * @param solutions
     * @return
     */
    final List<AbstractSolution> doReproduction(
            final List<AbstractSolution> solutions) {

        int original_solution_size = solutions.size();
        int nbr_weights = solutions.get(0).getWeightsP().size();

        //Select random parents for crossover
        while (solutions.size() < this.parameters.getPopulationSize()) {
            int dad = Utils.randomInteger(0, original_solution_size - 1);
            int mom;
            do {
                mom = Utils.randomInteger(0, original_solution_size - 1);
            } while (dad == mom);

            int cut_position = Utils.randomInteger(0, nbr_weights - 1);

            //beta is used to compute the new value at the cut position
            double beta = Math.random();
            //Reproduce parents to obtain 2 new children
            this.reproduce(solutions.get(dad),
                    solutions.get(mom),
                    solutions,
                    cut_position,
                    beta
            );
        }
        //Remove elements if the new population is bigger
        // than the initial population size.
        //It happened if the number of selected parents is odd
        while (solutions.size() > this.getParameters().getPopulationSize()) {
            solutions.remove(solutions.get(solutions.size() - 1));
        }
        return solutions;
    }

    /**
     * Two selected parents are combined to generate two new elements.
     * The method used is detailed in
     * https://gitlab.cylab.be/cylab/wowa-training/
     * blob/master/doc/TheContinuousGeneticAlgorithm.pdf p59
     *
     * @param dad
     * @param mom
     * @param solutions
     * @param cut_position
     * @param beta
     * @return
     */
    final void reproduce(final AbstractSolution dad,
                         final AbstractSolution mom,
                         final List<AbstractSolution> solutions,
                         final int cut_position,
                         final double beta) {

        double p_new1_w = dad.getWeightsW().get(cut_position) - beta
                * (dad.getWeightsW().get(cut_position)
                - mom.getWeightsW().get(cut_position));
        double p_new2_w = mom.getWeightsW().get(cut_position) + beta
                * (dad.getWeightsW().get(cut_position)
                - mom.getWeightsW().get(cut_position));

        double p_new1_p = dad.getWeightsP().get(cut_position) - beta
                * (dad.getWeightsP().get(cut_position)
                - mom.getWeightsP().get(cut_position));
        double p_new2_p = mom.getWeightsP().get(cut_position) + beta
                * (dad.getWeightsP().get(cut_position)
                - mom.getWeightsP().get(cut_position));

        AbstractSolution child1
                = this.factory.createSolutionObject(dad.weights_p.size());
        AbstractSolution child2
                = this.factory.createSolutionObject(dad.weights_p.size());

        for (int i = 0; i < cut_position; i++) {
            crossoverElements(child1, child2, dad, mom, i);
        }

        child1.getWeightsW().set(cut_position, p_new1_w);
        child2.getWeightsW().set(cut_position, p_new2_w);
        child1.getWeightsP().set(cut_position, p_new1_p);
        child2.getWeightsP().set(cut_position, p_new2_p);

        int nbre_weights = dad.getWeightsW().size();
        for (int i = cut_position + 1; i < nbre_weights; i++) {
            crossoverElements(child1, child2, mom, dad, i);
        }

        //It could be possible to obtain weight vector as [0,0,0,0,0]
        // in case of QuasiRandom initialization.
        //Check and correct vectors.
        if (this.getParameters().getPopulationInitializationMethod()
                == TrainerParameters.POPULATION_INITIALIZATION_QUASI_RANDOM) {
            this.checkAndCorrectNullWeightVector(child1);
            this.checkAndCorrectNullWeightVector(child2);
        }

        child1.normalize();
        child2.normalize();
        solutions.add(child1);
        solutions.add(child2);
    }

    private void crossoverElements(
            final AbstractSolution child1,
            final AbstractSolution child2,
            final AbstractSolution dad,
            final AbstractSolution mom,
            final int i) {
        child1.getWeightsW().set(i, dad.getWeightsW().get(i));
        child1.getWeightsP().set(i, dad.getWeightsP().get(i));

        child2.getWeightsW().set(i, mom.getWeightsW().get(i));
        child2.getWeightsP().set(i, mom.getWeightsP().get(i));
    }

    /**
     * Mutate random individuals with mutation rate probability.
     *
     * @param solutions
     * @return
     */
    final List<AbstractSolution> randomlyMutateGenes(
            final List<AbstractSolution> solutions) {

        double probability =
                (double) this.getParameters().getMutationRate() / 100;

        for (int i = 0; i < solutions.size(); i++) {
            if (i > 1) {
                solutions.get(i).randomlyMutateWithProbability(probability);
            }
        }

        return solutions;
    }

    /**
     * @return
     */
    final TrainerParameters getParameters() {
        return this.parameters;
    }

    /**
     * @param population_size
     * @param data
     * @param expected
     * @return
     */
    private List<AbstractSolution> generateInitialPopulationAndComputeDistances(
            final int population_size,
            final List<List<Double>> data,
            final List<Double> expected) {

        int number_of_weights = data.get(0).size();
        List<AbstractSolution> initial_population = new ArrayList<>();
        if (this.parameters.getPopulationInitializationMethod()
                == TrainerParameters.POPULATION_INITIALIZATION_RANDOM) {
            initial_population = this.generateInitialPopulation(
                    number_of_weights,
                    population_size
            );
        } else if (this.parameters.getPopulationInitializationMethod()
                == TrainerParameters.POPULATION_INITIALIZATION_QUASI_RANDOM) {
            initial_population = this.generateQuasiRandomInitialPopulation(
                    number_of_weights,
                    population_size);
        }
        return this.computeDistances(initial_population, data, expected);
    }

    /**
     * @param population
     * @param data
     * @param expected
     * @return
     */
    final List<AbstractSolution> performReproduction(
            final List<AbstractSolution> population,
            final List<List<Double>> data,
            final List<Double> expected) {

        List<AbstractSolution> parents = this.selectParents(population,
                this.getParameters().getNumberOfParents(),
                this.getParameters().getSelectionMethod());
        List<AbstractSolution> new_generation = this.doReproduction(parents);
        List<AbstractSolution> mutated
                = this.randomlyMutateGenes(new_generation);

        return this.computeDistances(mutated, data, expected);
    }

    /**
     * Correct weights vectors if it is null.
     * With QUASI_RANDOM initialization, it is possible to obtain a full null
     * weight vector after mating. The method check and correct that
     *
     * @param child AbstractSolution
     */
    private void checkAndCorrectNullWeightVector(final AbstractSolution child) {
        if (Utils.sumListElements(child.getWeightsW()) == 0) {
            int index = Utils.randomInteger(0, child.getWeightsW().size() - 1);
            List<Double> w = child.getWeightsW();
            w.set(index, 1.0);
            child.setWeightsW(w);
        }
        if (Utils.sumListElements(child.getWeightsP()) == 0) {
            int index = Utils.randomInteger(0, child.getWeightsP().size() - 1);
            List<Double> p = child.getWeightsP();
            p.set(index, 1.0);
            child.setWeightsP(p);
        }
    }
}
